Public Module Functions
    Public mode As Integer
    Public word_size As Integer
    Public dpoints As Integer
    Public sign As Integer
    Public func_key As Integer
    Public flag(3) As Boolean
    Public carry_bit As Boolean
    Public overflow As Boolean
    Private Calc_Form As Calc
    Private shift_window As Boolean
    Private sign_mask As New BigInt(32)
    ' The variables of the calculator
    Private reg(REG_SIZE - 1) As Double
    Private reg64(REG_SIZE - 1) As BigInt
    Private stack(STACK_SIZE - 1) As Double
    Private stack64(STACK_SIZE - 1) As BigInt
    Private lastx As Double
    Private lastx64 As New BigInt(32)
    Private op128 As New BigInt(128)

    ' Initialize the BigInt variable arrays
    Public Sub Initialize(ByVal f As Calc)
        Dim i As Integer

        For i = 0 To REG_SIZE - 1
            reg64(i) = New BigInt(32)
        Next
        For i = 0 To STACK_SIZE - 1
            stack64(i) = New BigInt(32)
        Next
        lastx64 = New BigInt(32)

        ' Build the initial sign mask
        sign_mask = BigInt.MakeSignMask(32, 32)

        ' Set a variable for the instance of the calling form
        Calc_Form = f
    End Sub

    Public Sub calc_key(ByVal key As Integer)
        ' Process the key strokes
        Dim i, j As Integer
        Dim store_it, input_overflow As Boolean
        Dim op As Double
        Dim op64 As New BigInt(word_size)
        Dim zero As New BigInt(word_size)
        Dim one As New BigInt(word_size, 1)
        Dim current As Double
        Dim current64 As New BigInt(word_size)
        Dim screen_buf, keystr As String

        Static input_buf As String
        Static pos, two_part, high_reg As Integer
        Static got_decimal, got_exponent, dont_push As Boolean

        ' String representation of the "naked" key
        keystr = Chr(key And 127)

        ' Turn off the F or G in the display
        If func_key > 0 Then
            func_key = 0
            DispFlags(DFLAG_FKEY, False)
            DispFlags(DFLAG_GKEY, False)
        End If

        ' Is this second half of a two part function?
        If two_part > 0 Then
            ' In the "real" calculator, you can't backspace over the beginning
            ' of a two-part function key... but I'm not that restrictive
            If key = K_BSP Or key = F_CPRE Then
                two_part = 0
                high_reg = 0
                If mode = FLOAT_MODE Then
                    Format_Display(stack(0))
                Else
                    Format_Display64(stack64(0), mode)
                End If
                Exit Sub
            End If

            ' What was the first half of this operation?
            Select Case two_part
                Case K_STO          ' store
                    If Chr(key) = "." Then
                        high_reg = 16
                        Exit Sub
                    End If
                    If keystr.IndexOfAny("0123456789abcdef") >= 0 Then
                        If key <= Asc("9") Then
                            i = Asc("0")
                        Else
                            i = Asc("a") + 10
                        End If
                        If mode = FLOAT_MODE Then
                            reg(key - i + high_reg) = stack(0)
                        Else
                            reg64(key - i + high_reg) = stack64(0)
                        End If
                    Else
                        Calc_Form.Display.Text = "Error 3 - Improper register number"
                        Beep()
                    End If
                    two_part = 0
                    high_reg = 0
                    Exit Sub
                Case K_RCL          ' recall
                    If Chr(key) = "." Then
                        high_reg = 16
                        Exit Sub
                    End If
                    If keystr.IndexOfAny("0123456789abcdef") >= 0 Then
                        If key <= Asc("9") Then
                            i = Asc("0")
                        Else
                            i = Asc("a") + 10
                        End If
                        If mode = FLOAT_MODE Then
                            current = reg(key - i + high_reg)
                            push(current)
                            Format_Display(stack(0))
                        Else
                            current64 = reg64(key - i + high_reg)
                            push64(current64)
                            Format_Display64(stack64(0), mode)
                        End If
                    Else
                        Calc_Form.Display.Text = "Error 3 - Improper register number"
                        Beep()
                    End If
                    two_part = 0
                    high_reg = 0
                    Exit Sub
                Case F_FLOAT        ' float digits
                    If keystr.IndexOfAny("0123456789.") >= 0 Then
                        dpoints = key - Asc("0")
                        ' If you are not already in float mode, then convert
                        ' everything from BigInt to double
                        If mode <> FLOAT_MODE Then
                            AllFrom64()
                        End If
                        mode = FLOAT_MODE
                        Format_Display(stack(0))

                        Calc_Form.MenuV_Float.Checked = True
                        Calc_Form.MenuV_Decimal.Checked = False
                        Calc_Form.MenuV_Hexadecimal.Checked = False
                        Calc_Form.MenuV_Octal.Checked = False
                        Calc_Form.MenuV_Binary.Checked = False
                    Else
                        Calc_Form.Display.Text = "Error 1 - Improper float number"
                        Beep()
                    End If
                    two_part = 0
                    Exit Sub
                Case G_SF           ' set flag
                    If keystr.IndexOfAny("012345") >= 0 Then
                        i = key - Asc("0")
                        Select Case i
                            Case 0 To 2
                                flag(i) = True
                            Case 3      ' leading zeros
                                flag(i) = True
                                If mode <> FLOAT_MODE Then
                                    Format_Display64(stack64(0), mode)
                                End If
                            Case 4
                                carry_bit = True
                                DispFlags(DFLAG_CARRYBIT, True)
                            Case 5
                                overflow = True
                                DispFlags(DFLAG_OVERFLOW, True)
                        End Select
                    Else
                        Calc_Form.Display.Text = "Error 1 - Improper flag number"
                        Beep()
                    End If
                    two_part = 0
                    Exit Sub
                Case G_CF           ' clear flag
                    If InStr("012345", Chr(key)) >= 0 Then
                        i = key - Asc("0")
                        Select Case i
                            Case 0 To 2
                                flag(i) = False
                            Case 3      ' leading zeros
                                flag(i) = False
                                If mode <> FLOAT_MODE Then
                                    Format_Display64(stack64(0), mode)
                                End If
                            Case 4
                                carry_bit = False
                                DispFlags(DFLAG_CARRYBIT, False)
                            Case 5
                                overflow = False
                                DispFlags(DFLAG_OVERFLOW, False)
                        End Select
                    Else
                        Calc_Form.Display.Text = "Error 1 - Improper flag number"
                        Beep()
                    End If
                    two_part = 0
                    Exit Sub
            End Select
        End If

        ' Is this numeric input?
        If (key < CTRL) Or (key = F_EEX) Then
            store_it = False
            input_overflow = False

            Select Case mode
                Case FLOAT_MODE
                    If keystr.IndexOfAny("0123456789.") >= 0 Then
                        store_it = True
                    End If
                    ' Float mode allows more than just digits
                    Select Case key
                        Case K_DP   ' decimal point
                            If (got_decimal Or got_exponent) Then
                                Beep()
                                Exit Sub
                            End If
                            got_decimal = True
                        Case F_EEX  ' exponent
                            If (pos = 0 Or got_exponent) Then
                                Beep()
                                Exit Sub
                            End If
                            got_exponent = True
                            key = Asc("e")
                            store_it = True
                        Case K_CHS  ' change sign
                            If pos > 0 Then
                                store_it = True
                                ' Change sign of exponent?
                                i = input_buf.IndexOf("e")
                                If i >= 0 Then
                                    If input_buf.Substring(i + 1, 1) = "-" Then
                                        input_buf = input_buf.Substring(0, i + 1) & input_buf.Substring(i + 2)
                                        pos = pos - 1
                                    Else
                                        input_buf = input_buf.Substring(0, i + 1) & "-" & input_buf.Substring(i + 1)
                                        pos = pos + 1
                                    End If
                                Else
                                    ' OK, just toggle the sign
                                    If input_buf.Substring(0, 1) = "-" Then
                                        input_buf = input_buf.Substring(1)
                                        pos = pos - 1
                                    Else
                                        input_buf = "-" & input_buf
                                        pos = pos + 1
                                    End If
                                End If
                            End If
                    End Select
                    If pos >= MAX_FLOAT_DIGITS Then
                        input_overflow = True
                    End If
                Case DEC_MODE
                    If keystr.IndexOfAny("0123456789") >= 0 Then
                        store_it = True
                    End If
                    If pos >= MAX_DECIMAL_DIGITS Then
                        input_overflow = True
                    End If
                Case HEX_MODE
                    If keystr.IndexOfAny("0123456789abcdef") >= 0 Then
                        store_it = True
                    End If
                    If pos >= MAX_HEX_DIGITS Then
                        input_overflow = True
                    End If
                Case OCT_MODE
                    If keystr.IndexOfAny("01234567") >= 0 Then
                        store_it = True
                    End If
                    If pos >= MAX_OCTAL_DIGITS Then
                        input_overflow = True
                    End If
                Case BIN_MODE
                    If keystr.IndexOfAny("01") >= 0 Then
                        store_it = True
                    End If
                    If pos >= MAX_BINARY_DIGITS Then
                        input_overflow = True
                    End If
            End Select
        End If

        ' Store numeric input
        If store_it = True Then
            If input_overflow = True Then
                Beep()
                Exit Sub
            End If
            ' Placement of the "-" is handled by K_CHS
            If (key <> K_CHS) Then
                pos = pos + 1
                input_buf = input_buf & Chr(key)
            End If
            Raw_Display(input_buf)
            Exit Sub
        End If

        ' Process "pseudo" operators (things that don't force the
        ' conversion of pending input)
        Select Case key
            Case K_BSP
                If pos > 0 Then
                    input_buf = input_buf.Substring(0, pos - 1)
                    pos = pos - 1
                    Raw_Display(input_buf)
                Else    ' same as clear x
                    If mode = FLOAT_MODE Then
                        stack(0) = 0.0#
                        Format_Display(stack(0))
                    Else
                        stack64(0) = zero
                        Format_Display64(stack64(0), mode)
                    End If
                    dont_push = True
                End If
                Exit Sub
            Case K_FKEY
                func_key = CTRL
                DispFlags(DFLAG_FKEY, True)
                Exit Sub
            Case K_GKEY
                func_key = ALT
                DispFlags(DFLAG_GKEY, True)
                Exit Sub
        End Select

        ' If we get this far, we have pushed a function key with numeric
        ' input still pending on the screen.
        If pos > 0 Then
            If mode = FLOAT_MODE Then
                current = Convert_Input(input_buf)
                ' This is a little weird... if the last operation was K_ENTER
                ' or G_CLX then replace (don't push) the value on the stack
                If dont_push Then
                    stack(0) = current
                    dont_push = False
                Else
                    push(current)
                End If
            Else
                current64 = Convert_Input64(input_buf)
                If dont_push Then
                    stack64(0) = current64
                    dont_push = False
                Else
                    push64(current64)
                End If
            End If
            pos = 0
            got_decimal = False
            got_exponent = False
            input_buf = ""
        End If

        ' *
        ' * Process the function keys
        ' *
        Select Case key
            Case 0                  ' just update the display
            Case K_DIV              ' division
                If mode = FLOAT_MODE Then
                    lastx = stack(0)
                    op = pop()
                    If op = 0 Then
                        Calc_Form.Display.Text = "Error 0 - Improper Math Operation"
                        Beep()
                        Exit Sub
                    End If
                    current = pop() / op
                    push(current)
                Else
                    lastx64 = stack64(0)
                    op64 = pop64()
                    If BigInt.IsZero(op64) Then
                        Calc_Form.Display.Text = "Error 0 - Improper Math Operation"
                        Beep()
                        Exit Sub
                    End If
                    current64 = BigInt.Divide(pop64(), op64)
                    DoFlags(current64)
                    push64(current64)
                End If
            Case K_GSB              ' goto subroutine
                Calc_Form.Display.Text = "GSB not implemented"
                Exit Sub
            Case K_GTO              ' goto label
                Calc_Form.Display.Text = "GTO not implemented"
                Exit Sub
            Case K_HEX              ' hexadecimal mode
                Calc_Form.MenuV_Float.Checked = False
                Calc_Form.MenuV_Decimal.Checked = False
                Calc_Form.MenuV_Hexadecimal.Checked = True
                Calc_Form.MenuV_Octal.Checked = False
                Calc_Form.MenuV_Binary.Checked = False
                If mode = FLOAT_MODE Then
                    AllTo64()
                End If
                mode = HEX_MODE
            Case K_DEC              ' decimal mode
                Calc_Form.MenuV_Float.Checked = False
                Calc_Form.MenuV_Decimal.Checked = True
                Calc_Form.MenuV_Hexadecimal.Checked = False
                Calc_Form.MenuV_Octal.Checked = False
                Calc_Form.MenuV_Binary.Checked = False
                If mode = FLOAT_MODE Then
                    AllTo64()
                End If
                mode = DEC_MODE
            Case K_OCT              ' octal mode
                Calc_Form.MenuV_Float.Checked = False
                Calc_Form.MenuV_Decimal.Checked = False
                Calc_Form.MenuV_Hexadecimal.Checked = False
                Calc_Form.MenuV_Octal.Checked = True
                Calc_Form.MenuV_Binary.Checked = False
                If mode = FLOAT_MODE Then
                    AllTo64()
                End If
                mode = OCT_MODE
            Case K_BIN              ' binary mode
                Calc_Form.MenuV_Float.Checked = False
                Calc_Form.MenuV_Decimal.Checked = False
                Calc_Form.MenuV_Hexadecimal.Checked = False
                Calc_Form.MenuV_Octal.Checked = False
                Calc_Form.MenuV_Binary.Checked = True
                If mode = FLOAT_MODE Then
                    AllTo64()
                End If
                mode = BIN_MODE
            Case K_MULT             ' multiplication
                If mode = FLOAT_MODE Then
                    lastx = stack(0)
                    current = pop() * pop() ' fizz, fizz, oh what a relief it is
                    push(current)
                Else
                    lastx64 = stack64(0)
                    current64 = BigInt.Multiply(pop64(), pop64())
                    DoFlags(current64)
                    push64(current64)
                End If
            Case K_RS               ' run/stop
                Calc_Form.Display.Text = "RS not implemented"
                Exit Sub
            Case K_SST              ' single step
                Calc_Form.Display.Text = "SST not implemented"
                Exit Sub
            Case K_ROL              ' roll the stack
                If mode = FLOAT_MODE Then
                    op = stack(0)
                    For i = 0 To STACK_SIZE - 2
                        stack(i) = stack(i + 1)
                    Next
                    stack(STACK_SIZE - 1) = op
                Else
                    op64 = stack64(0)
                    For i = 0 To STACK_SIZE - 2
                        stack64(i) = stack64(i + 1)
                    Next
                    stack64(STACK_SIZE - 1) = op64
                End If
            Case K_XY               ' exchange X and Y
                If mode = FLOAT_MODE Then
                    op = pop()
                    current = pop()
                    push(op)
                    push(current)
                Else
                    op64 = pop64()
                    current64 = pop64()
                    push64(op64)
                    push64(current64)
                End If
            Case K_ENTER            ' enter key
                dont_push = True
                If mode = FLOAT_MODE Then
                    push(current)
                Else
                    push64(current64)
                End If
            Case K_MINUS            ' subtraction
                If mode = FLOAT_MODE Then
                    lastx = stack(0)
                    op = pop()
                    current = pop() - op
                    push(current)
                Else
                    lastx64 = stack64(0)
                    op64 = pop64()
                    current64 = BigInt.Subtract(pop64(), op64)
                    DoFlags(current64)
                    push64(current64)
                End If
            Case K_STO              ' store into a register
                two_part = K_STO
            Case K_RCL              ' recall from a register
                two_part = K_RCL
            Case K_CHS              ' change sign
                If mode = FLOAT_MODE Then
                    current = pop() * -1.0#
                    push(current)
                Else
                    lastx64 = stack64(0)
                    Select Case sign
                        Case SIGN_UNSIGNED
                            current64 = BigInt.Add(BigInt.LogicalNot(pop64()), one)
                            ' Not really an error, just set the overflow flag
                            overflow = True
                            DispFlags(DFLAG_OVERFLOW, True)
                        Case SIGN_1S_COMPLEMENT
                            current64 = BigInt.LogicalNot(pop64())
                        Case SIGN_2S_COMPLEMENT
                            current64 = BigInt.Add(BigInt.LogicalNot(pop64()), one)
                    End Select
                    push64(current64)
                End If
            Case K_PLUS             ' addition
                If mode = FLOAT_MODE Then
                    lastx = stack(0)
                    current = pop() + pop()
                    push(current)
                Else
                    lastx64 = stack64(0)
                    current64 = BigInt.Add(pop64(), pop64())
                    DoFlags(current64)
                    push64(current64)
                End If
                '
                ' And now for the yellow/f (ctrl) functions
                '
            Case F_SL               ' shift left
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                current64 = BigInt.ShiftLeft(pop64())
                DoFlags(current64)
                push64(current64)
            Case F_SR               ' shift right
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                current64 = BigInt.ShiftRight(pop64())
                DoFlags(current64)
                push64(current64)
            Case F_RL               ' rotate left
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                current64 = BigInt.ShiftLeft(pop64())
                DoFlags(current64)
                If carry_bit Then
                    current64 = BigInt.LogicalOr(current64, one)
                End If
                push64(current64)
            Case F_RR               ' rotate right
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                current64 = BigInt.ShiftRight(pop64())
                DoFlags(current64)
                If carry_bit Then
                    current64 = BigInt.LogicalOr(current64, sign_mask)
                End If
                push64(current64)
            Case F_RLN              ' rotate left "n" times
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                ' The number of times to do it
                j = pop64().ToInteger
                If j < 1 Or j > word_size Then
                    Calc_Form.Display.Text = "Error 2 - Improper Bit Number"
                    Beep()
                    Exit Sub
                End If
                current64 = pop64()
                For i = 1 To j
                    current64 = BigInt.ShiftLeft(current64)
                    carry_bit = current64.CarryBit()
                    If carry_bit Then
                        current64 = BigInt.LogicalOr(current64, one)
                    End If
                Next
                ' Can't use doflags, since we may have lost last carry flag
                If carry_bit Then
                    DispFlags(DFLAG_CARRYBIT, True)
                Else
                    DispFlags(DFLAG_CARRYBIT, False)
                End If
                push64(current64)
            Case F_RRN              ' rotate right "n" times
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                ' The number of times to do it
                j = pop64().ToInteger
                If j < 1 Or j > word_size Then
                    Calc_Form.Display.Text = "Error 2 - Improper Bit Number"
                    Beep()
                    Exit Sub
                End If
                current64 = pop64()
                For i = 1 To j
                    current64 = BigInt.ShiftRight(current64)
                    carry_bit = current64.CarryBit()
                    If carry_bit Then
                        current64 = BigInt.LogicalOr(current64, sign_mask)
                    End If
                Next
                ' Can't use doflags, since we may have lost last carry flag
                If carry_bit Then
                    DispFlags(DFLAG_CARRYBIT, True)
                Else
                    DispFlags(DFLAG_CARRYBIT, False)
                End If
                push64(current64)
            Case F_MASKL            ' mask left
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                i = pop64().ToInteger
                If i < 1 Or i > word_size Then
                    Calc_Form.Display.Text = "Error 2 - Improper Bit Number"
                    Beep()
                    Exit Sub
                End If
                current64 = BigInt.LogicalNot(BigInt.MakeWordMask(word_size - i, word_size))
                push64(current64)
            Case F_MASKR            ' mask right
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                i = pop64().ToInteger()
                If i < 1 Or i > word_size Then
                    Calc_Form.Display.Text = "Error 2 - Improper Bit Number"
                    Beep()
                    Exit Sub
                End If
                current64 = BigInt.MakeWordMask(i, word_size)
                push64(current64)
            Case F_RMD              ' remainder (modulus)
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                op64 = pop64()
                If BigInt.IsZero(op64) Then
                    Calc_Form.Display.Text = "Error 0 - Improper Math Operation"
                    Beep()
                    Exit Sub
                End If
                current64 = BigInt.Modulus(pop64(), op64)
                DoFlags(current64)
                push64(current64)
            Case F_XOR              ' exclusive or
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                op64 = pop64()
                current64 = BigInt.LogicalXor(pop64(), op64)
                push64(current64)
            Case F_XIND             ' exchange x with contents of index
                Calc_Form.Display.Text = "x:(i) not implemented"
                Exit Sub
            Case F_XI               ' exchange x with index
                Calc_Form.Display.Text = "x:i not implemented"
                Exit Sub
            Case F_SHEX             ' show hex
                op64 = BigInt.Assign(stack(0), word_size)
                Format_Display64(op64, HEX_MODE)
                Exit Sub
            Case F_SDEC             ' show decimal
                op64 = BigInt.Assign(stack(0), word_size)
                Format_Display64(op64, DEC_MODE)
                Exit Sub
            Case F_SOCT             ' show octal
                op64 = BigInt.Assign(stack(0), word_size)
                Format_Display64(op64, OCT_MODE)
                Exit Sub
            Case F_SBIN             ' show binary
                op64 = BigInt.Assign(stack(0), word_size)
                Format_Display64(op64, BIN_MODE)
                Exit Sub
            Case F_SB               ' set bit
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                i = pop64().ToInteger
                If i < 0 Or i > word_size Then
                    Calc_Form.Display.Text = "Error 2 - Improper Bit Number"
                    Beep()
                    Exit Sub
                End If
                op64 = BigInt.MakeSignMask(i + 1, word_size)
                current64 = BigInt.LogicalOr(pop64(), op64)
                push64(current64)
            Case F_CB               ' clear bit
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                i = pop64().ToInteger
                If i < 0 Or i > word_size Then
                    Calc_Form.Display.Text = "Error 2 - Improper Bit Number"
                    Beep()
                    Exit Sub
                End If
                op64 = BigInt.LogicalNot(BigInt.MakeSignMask(i + 1, word_size))
                current64 = BigInt.LogicalAnd(pop64(), op64)
                push64(current64)
            Case F_BSET             ' is bit set?
                Calc_Form.Display.Text = "B? not implemented"
                Exit Sub
            Case F_AND              ' logical and
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                current64 = BigInt.LogicalAnd(pop64(), pop64())
                push64(current64)
            Case F_IND              ' contents of index
                Calc_Form.Display.Text = "(i) not implemented"
                Exit Sub
            Case F_I                ' index
                Calc_Form.Display.Text = "I not implemented"
                Exit Sub
            Case F_CPGRM            ' clear program
                Calc_Form.Display.Text = "CPGRM not implemented"
                Exit Sub
            Case F_CREG             ' clear registers
                If mode = FLOAT_MODE Then
                    For i = 0 To REG_SIZE - 1
                        reg(i) = 0.0#
                    Next
                Else
                    For i = 0 To REG_SIZE - 1
                        reg64(i) = zero
                    Next
                End If
                ' F_PREFIX is handled elsewhere
            Case F_WIN              ' set window number
                ' I make a departure from the real calculator.  I silently
                ' ignore the WINDOW command, and use the "<" and ">" operators
                ' to move the entire 32 bit window.
            Case F_S1S              ' 1s complement
                sign = SIGN_1S_COMPLEMENT
                Calc_Form.MenuO_1sComplement.Checked = True
                Calc_Form.MenuO_2sComplement.Checked = False
                Calc_Form.MenuO_Unsigned.Checked = False
            Case F_S2S              ' 2s complement
                sign = SIGN_2S_COMPLEMENT
                Calc_Form.MenuO_1sComplement.Checked = False
                Calc_Form.MenuO_2sComplement.Checked = True
                Calc_Form.MenuO_Unsigned.Checked = False
            Case F_SUNS             ' unsigned complement
                sign = SIGN_UNSIGNED
                Calc_Form.MenuO_1sComplement.Checked = False
                Calc_Form.MenuO_2sComplement.Checked = False
                Calc_Form.MenuO_Unsigned.Checked = True
            Case F_NOT              ' logical not
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                current64 = BigInt.LogicalNot(pop64())
                push64(current64)
            Case F_WSIZE            ' word size
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                i = pop64().ToInteger
                If (i < 1 Or i > 64) Then
                    Calc_Form.Display.Text = "Error 2 - Improper Bit Number"
                    Beep()
                    Exit Sub
                End If
                sign_mask = BigInt.MakeSignMask(i, i)
                ' Convert all existing numbers to fit the new word_size
                If word_size <> i Then
                    ResizeAll(i)
                End If
                word_size = i
                ' Turn off all the checks (could be an "non-standard" size)
                Calc_Form.MenuO_8bit.Checked = False
                Calc_Form.MenuO_16bit.Checked = False
                Calc_Form.MenuO_32bit.Checked = False
                Calc_Form.MenuO_64bit.Checked = False
                Select Case word_size
                    Case 8
                        Calc_Form.MenuO_8bit.Checked = True
                    Case 16
                        Calc_Form.MenuO_16bit.Checked = True
                    Case 32
                        Calc_Form.MenuO_32bit.Checked = True
                    Case 64
                        Calc_Form.MenuO_64bit.Checked = True
                End Select
            Case F_FLOAT            ' floating point mode
                two_part = F_FLOAT
                Exit Sub
            Case F_MEM              ' display memory
                Calc_Form.Display.Text = "MEM not implemented"
                Exit Sub
            Case F_STAT             ' display status
                screen_buf = sign & "-" & Format(word_size, "00") & "-"
                ' The user flags
                For i = 3 To 0 Step -1
                    If flag(i) = True Then
                        screen_buf = screen_buf & "1"
                    Else
                        screen_buf = screen_buf & "0"
                    End If
                Next
                Calc_Form.Display.Text = screen_buf
                Exit Sub
                ' F_EEX is handled elsewhere
            Case F_OR               ' logical or
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                current64 = BigInt.LogicalOr(pop64(), pop64())
                push64(current64)
                '
                ' And now for the blue/g (alt) functions
                '
            Case G_LJ               ' left justify
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                op64 = pop64()
                ' Short cut for zero
                If BigInt.IsZero(op64) Then
                    current64 = op64
                    push64(zero)
                Else
                    ' Where is first bit?
                    i = word_size - op64.MaxBit() - 1
                    ' Now shift left that many times
                    current64 = op64
                    For j = 1 To i
                        current64 = BigInt.ShiftLeft(current64)
                    Next
                End If
                push64(current64)
                op64 = BigInt.Assign(i, word_size)
                push64(op64)
            Case G_ASR              ' arithmetic shift right
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                op64 = pop64()
                If op64.IsNegative() Then
                    op64 = BigInt.ShiftRight(op64)
                    DoFlags(op64)
                    current64 = BigInt.LogicalOr(op64, sign_mask)
                Else
                    current64 = BigInt.ShiftRight(op64)
                    DoFlags(current64)
                End If
                push64(current64)
            Case G_RLC              ' rotate left with carry
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                ' The previous carry bit
                If carry_bit Then
                    op64 = BigInt.ShiftLeft(pop64())
                    DoFlags(op64)
                    current64 = BigInt.LogicalOr(op64, one)
                Else
                    current64 = BigInt.ShiftLeft(pop64())
                    DoFlags(current64)
                End If
                push64(current64)
            Case G_RRC              ' rotate right with carry
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                ' The previous carry bit
                If carry_bit Then
                    op64 = BigInt.ShiftRight(pop64())
                    DoFlags(op64)
                    current64 = BigInt.LogicalOr(op64, sign_mask)
                Else
                    current64 = BigInt.ShiftRight(pop64())
                    DoFlags(current64)
                End If
                push64(current64)
            Case G_RLCN             ' rotate left with carry "n" times
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                ' The number of times to do it
                j = pop64().ToInteger
                If j < 1 Or j > word_size Then
                    Calc_Form.Display.Text = "Error 2 - Improper Bit Number"
                    Beep()
                    Exit Sub
                End If
                current64 = pop64()
                For i = 1 To j
                    ' Start with previous carry bit
                    If carry_bit Then
                        op64 = BigInt.ShiftLeft(current64)
                        carry_bit = op64.CarryBit()
                        current64 = BigInt.LogicalOr(op64, one)
                    Else
                        current64 = BigInt.ShiftLeft(current64)
                        carry_bit = current64.CarryBit()
                    End If
                Next
                ' Can't use doflags, since we may have lost last carry flag
                If carry_bit Then
                    DispFlags(DFLAG_CARRYBIT, True)
                Else
                    DispFlags(DFLAG_CARRYBIT, False)
                End If
                push64(current64)
            Case G_RRCN             ' rotate right with carry "n" times
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                ' The number of times to do it
                j = pop64().ToInteger
                If j < 1 Or j > word_size Then
                    Calc_Form.Display.Text = "Error 2 - Improper Bit Number"
                    Beep()
                    Exit Sub
                End If
                current64 = pop64()
                For i = 1 To j
                    ' Start with previous carry bit
                    If carry_bit Then
                        op64 = BigInt.ShiftRight(current64)
                        carry_bit = op64.CarryBit()
                        current64 = BigInt.LogicalOr(op64, sign_mask)
                    Else
                        current64 = BigInt.ShiftRight(current64)
                        carry_bit = current64.CarryBit()
                    End If
                Next
                ' Can't use doflags, since we may have lost last carry flag
                If carry_bit Then
                    DispFlags(DFLAG_CARRYBIT, True)
                Else
                    DispFlags(DFLAG_CARRYBIT, False)
                End If
                push64(current64)
            Case G_BITS             ' sum the number of bits
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                op64 = pop64()
                i = op64.SumBits()
                current64 = BigInt.Assign(i, word_size)
                push64(current64)
            Case G_ABS              ' absolute value
                If mode = FLOAT_MODE Then
                    lastx = stack(0)
                    current = Math.Abs(pop())
                    push(current)
                Else
                    lastx64 = stack64(0)
                    op64 = pop64()
                    If op64.IsNegative() Then
                        Select Case sign
                            Case SIGN_UNSIGNED
                                current64 = op64
                            Case SIGN_1S_COMPLEMENT
                                current64 = BigInt.LogicalNot(op64)
                            Case SIGN_2S_COMPLEMENT
                                current64 = BigInt.Add(BigInt.LogicalNot(op64), one)
                        End Select
                    Else
                        current64 = op64
                    End If
                    push64(current64)
                End If
            Case G_DBLR             ' double remainder
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                op64 = pop64()
                current64 = pop64()
                op128.SetSize(word_size * 2)
                op128.Combine(pop64(), current64)
                current64 = BigInt.Modulus(op128, op64)
                ' Have to return it back to word_size
                current64.SetSize(word_size)
                push64(current64)
            Case G_DDIV             ' double divide
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                op64 = pop64()
                current64 = pop64()
                op128.SetSize(word_size * 2)
                op128.Combine(pop64(), current64)
                current64 = BigInt.Divide(op128, op64)
                ' Have to return it back to word_size
                current64.SetSize(word_size)
                push64(current64)
            Case G_RTN              ' return
                Calc_Form.Display.Text = "RTN not implemented"
                Exit Sub
            Case G_LBL              ' label
                Calc_Form.Display.Text = "LBL not implemented"
                Exit Sub
            Case G_DSZ              ' decrement skip on zero
                Calc_Form.Display.Text = "DSZ not implemented"
                Exit Sub
            Case G_ISZ              ' increment skip on zero
                Calc_Form.Display.Text = "ISZ not implemented"
                Exit Sub
            Case G_SQRT             ' square root
                If mode = FLOAT_MODE Then
                    lastx = stack(0)
                    op = pop()
                    If op < 0 Then
                        Calc_Form.Display.Text = "Error 0 - Improper Math Operation"
                        Beep()
                        Exit Sub
                    End If
                    current = Math.Sqrt(op)
                    push(current)
                Else
                    lastx64 = stack64(0)
                    op64 = pop64()
                    ' The real calculator tests for a negative number in the
                    ' 1's comp and 2's comp mode.  I don't
                    If BigInt.IsZero(op64) Then
                        Calc_Form.Display.Text = "Error 0 - Improper Math Operation"
                        Beep()
                        Exit Sub
                    End If
                    current64 = BigInt.SquareRoot(op64)
                    push64(current64)
                End If
            Case G_INV              ' inverse
                If mode <> FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx = stack(0)
                op = pop()
                If op = 0 Then
                    Calc_Form.Display.Text = "Error 0 - Improper Math Operation"
                    Beep()
                    Exit Sub
                End If
                current = 1 / op
                push(current)
            Case G_SF               ' set flag
                two_part = G_SF
            Case G_CF               ' clear flag
                two_part = G_CF
            Case G_FSET             ' is flag set?
                Calc_Form.Display.Text = "F? not implemented"
                Exit Sub
            Case G_DMUL             ' double multiply
                If mode = FLOAT_MODE Then
                    Beep()
                    Exit Sub
                End If
                lastx64 = stack64(0)
                op128 = BigInt.Multiply(stack64(0), stack64(1), word_size * 2)
                ' Double multiply returns two values... directly to the stack
                op128.Split(stack64(0), stack64(1))
            Case G_PR               ' program / run mode
                Calc_Form.Display.Text = "PR not implemented"
                Exit Sub
            Case G_BST              ' back step
                Calc_Form.Display.Text = "BST not implemented"
                Exit Sub
            Case G_RUP              ' roll stack up
                If mode = FLOAT_MODE Then
                    op = stack(STACK_SIZE - 1)
                    For i = STACK_SIZE - 1 To 1 Step -1
                        stack(i) = stack(i - 1)
                    Next
                    stack(0) = op
                Else
                    op64 = stack64(STACK_SIZE - 1)
                    For i = STACK_SIZE - 1 To 1 Step -1
                        stack64(i) = stack64(i - 1)
                    Next
                    stack64(0) = op64
                End If
            Case G_PSE              ' pause
                Calc_Form.Display.Text = "PSE not implemented"
                Exit Sub
            Case G_CLX              ' clear x
                If mode = FLOAT_MODE Then
                    stack(0) = 0.0#
                Else
                    stack64(0) = zero
                End If
                dont_push = True
            Case G_LSTX             ' last x
                If mode = FLOAT_MODE Then
                    push(lastx)
                Else
                    push64(lastx64)
                End If
            Case G_XGEY             ' x >= y
                Calc_Form.Display.Text = "X>=Y not implemented"
                Exit Sub
            Case G_XL0              ' x < 0
                Calc_Form.Display.Text = "X<0 not implemented"
                Exit Sub
            Case G_XGY              ' x > y
                Calc_Form.Display.Text = "X>Y not implemented"
                Exit Sub
            Case G_XG0              ' x > 0
                Calc_Form.Display.Text = "X>0 not implemented"
                Exit Sub
            Case G_WLEFT            ' move window to the left
                ' This is significantly different from the real calculator.
                ' Instead of moving the window 1 bit at a time, it gets
                ' moved 32 bits.  More like the WINDOW function.
                If mode <> BIN_MODE Then
                    Beep()
                    Exit Sub
                End If
                If word_size <= 32 Then
                    ' Ignore if nothing to do
                    Exit Sub
                End If
                shift_window = True
            Case G_WRIGHT           ' move window to the right
                ' This is significantly different from the real calculator.
                ' Instead of moving the window 1 bit at a time, it gets
                ' moved 32 bits.  More like the WINDOW function.
                If mode <> BIN_MODE Then
                    Beep()
                    Exit Sub
                End If
                ' Actually, since "right window" is the default, we don't
                ' have to do anything!
            Case G_XNEY             ' x <> y
                Calc_Form.Display.Text = "X<>Y not implemented"
                Exit Sub
            Case G_XNE0             ' x <> 0
                Calc_Form.Display.Text = "X<>0 not implemented"
                Exit Sub
            Case G_XEY              ' x = y
                Calc_Form.Display.Text = "X=Y not implemented"
                Exit Sub
            Case G_XE0              ' x = 0
                Calc_Form.Display.Text = "X=0 not implemented"
                Exit Sub
            Case Else
                Beep()
        End Select

        ' Display the answer
        If mode = FLOAT_MODE Then
            Format_Display(stack(0))
        Else
            Format_Display64(stack64(0), mode)
        End If
    End Sub

    ' *
    ' * Stack operations
    ' *

    ' Push a value onto the stack.  If full, the top of the stack is lost
    Public Sub push(ByVal val As Double)
        Dim i As Integer

        For i = STACK_SIZE - 1 To 1 Step -1
            stack(i) = stack(i - 1)
        Next
        stack(0) = val
    End Sub

    ' Push a BigInt value onto the stack.  If full, the top of the stack
    ' is lost
    Public Sub push64(ByVal val As BigInt)
        Dim i As Integer

        For i = STACK_SIZE - 1 To 1 Step -1
            stack64(i) = stack64(i - 1)
        Next
        stack64(0) = val
    End Sub

    ' Pop a Double value off of the stack.  The top value is preserved
    Private Function pop() As Double
        Dim f As Double
        Dim i As Integer

        f = stack(0)
        lastx = f
        For i = 0 To STACK_SIZE - 2
            stack(i) = stack(i + 1)
        Next
        Return f
    End Function

    ' Pop a BigInt value off of the stack.  The top value is preserved
    Private Function pop64() As BigInt
        Dim f As BigInt
        Dim i As Integer

        f = stack64(0)
        lastx64 = f
        For i = 0 To STACK_SIZE - 2
            stack64(i) = stack64(i + 1)
        Next
        Return f
    End Function

    '*
    '* Input conversion routines
    '*

    ' Convert ASCII input to a double
    Private Function Convert_Input(ByVal str As String) As Double
        Dim f As Double

        Try
            f = CDbl(str)
        Catch
        End Try

        Return f
    End Function

    ' Convert ASCII input into a BigInt
    Private Function Convert_Input64(ByVal str As String) As BigInt
        Dim f As BigInt

        f = Nothing
        Select Case mode
            Case HEX_MODE
                f = BigInt.Assign("&H" & str, word_size)
            Case DEC_MODE
                f = BigInt.Assign(str, word_size)
            Case OCT_MODE
                f = BigInt.Assign("&O" & str, word_size)
            Case BIN_MODE
                ' Zap the byte spacers (if any)
                f = BigInt.Assign("&B" & str.Replace(" ", ""), word_size)
        End Select

        Return f
    End Function

    ' Convert text pasted into the calculator's display
    Public Sub ConvertFromPaste(ByVal buf As String)
        Dim f As Double
        Dim x As BigInt

        If mode = FLOAT_MODE Then
            f = Convert_Input(buf)
            push(f)
        Else
            x = Convert_Input64(buf)
            push64(x)
        End If
    End Sub

    ' *
    ' * A bunch of Display functions
    ' *

    ' Format a double for the display
    Private Sub Format_Display(ByVal f As Double)
        Dim i As Integer
        Dim buf As String
        Dim point_mask As String
        Dim screen_buf As String

        ' Scientific notation?
        If dpoints < 0 Then
            ' Note: dpoints is negative because I interpreted the "."
            ' character as an ASCII digit.
            point_mask = "0.000000000e-0"
            screen_buf = Format(f, point_mask)
        Else
            point_mask = "0."
            For i = 1 To dpoints
                point_mask = point_mask & "0"
            Next
            buf = Format(f, point_mask)
            ' Run out of screen space?  Try scientific notation
            If (buf.Length >= SCREEN_SIZE) Or ((buf = point_mask) And (f <> 0.0#)) Then
                point_mask = point_mask & "e-0"
                screen_buf = Format(f, point_mask)
            Else
                screen_buf = buf
            End If
        End If
        Calc_Form.Display.Text = screen_buf

        ' Every once and a while, we should check to see if there is
        ' something on the clipboard with the correct format
        If Clipboard.GetDataObject.GetDataPresent(DataFormats.Text) Then
            Calc_Form.MenuE_Paste.Enabled = True
        Else
            Calc_Form.MenuE_Paste.Enabled = False
        End If
    End Sub

    ' Format a BigInt for the display
    Private Sub Format_Display64(ByVal f As BigInt, ByVal mode As Integer)
        Dim buf As String
        Dim start As Integer
        Dim screen_buf As String = ""

        ' My version of flag(3) is to control leading zeros in the hex, dec,
        ' and oct modes (but not in the bin mode!).  This is exactly opposite
        ' of the real calculator
        Select Case mode
            Case HEX_MODE
                buf = f.ToStringHex()
                If flag(3) = False Then
                    buf = TrimZeros(buf)
                End If
                screen_buf = Space(SCREEN_SIZE) & buf & " h "
                screen_buf = screen_buf.Substring(screen_buf.Length - SCREEN_SIZE, SCREEN_SIZE)
            Case OCT_MODE
                buf = f.ToStringOct()
                If flag(3) = False Then
                    buf = TrimZeros(buf)
                End If
                screen_buf = Space(SCREEN_SIZE) & buf & " o "
                screen_buf = screen_buf.Substring(screen_buf.Length - SCREEN_SIZE, SCREEN_SIZE)
            Case DEC_MODE
                buf = f.ToStringDec(sign)
                If flag(3) = False Then
                    buf = TrimZeros(buf)
                End If
                screen_buf = Space(SCREEN_SIZE) & buf & " d "
                screen_buf = screen_buf.Substring(screen_buf.Length - SCREEN_SIZE, SCREEN_SIZE)
            Case BIN_MODE
                ' The binary mode is the only mode which can't always fit the
                ' answer on the screen.  So we sometimes display only half of
                ' the answer.
                buf = f.ToStringBin(True)
                If shift_window Then
                    start = SCREEN_SIZE - word_size - ((word_size - 32) / 8) + 31
                    ' Ignore the "trim leading zeros" in binary mode.
                    screen_buf = Space(SCREEN_SIZE) & buf.Substring(start) & " b."
                    screen_buf = screen_buf.Substring(screen_buf.Length - SCREEN_SIZE, SCREEN_SIZE)
                    shift_window = False
                Else
                    If word_size <= 32 Then
                        screen_buf = Space(SCREEN_SIZE) & buf & " b "
                        screen_buf = screen_buf.Substring(screen_buf.Length - SCREEN_SIZE, SCREEN_SIZE)
                    Else
                        screen_buf = buf & " .b"
                        screen_buf = screen_buf.Substring(screen_buf.Length - SCREEN_SIZE, SCREEN_SIZE)
                    End If
                End If
        End Select
        Calc_Form.Display.Text = screen_buf

        ' Every once and a while, we should check to see if there is
        ' something on the clipboard with the correct format
        If Clipboard.GetDataObject.GetDataPresent(DataFormats.Text) Then
            Calc_Form.MenuE_Paste.Enabled = True
        Else
            Calc_Form.MenuE_Paste.Enabled = False
        End If
    End Sub

    ' Format user's (raw) input for the display
    Private Sub Raw_Display(ByVal input_buf As String)
        Dim buf As String = ""

        Select Case mode
            Case FLOAT_MODE
                buf = input_buf & Space(SCREEN_SIZE)
                buf = buf.Substring(0, SCREEN_SIZE)
            Case HEX_MODE
                buf = Space(SCREEN_SIZE) & input_buf & " h "
                buf = buf.Substring(buf.Length - SCREEN_SIZE, SCREEN_SIZE)
            Case OCT_MODE
                buf = Space(SCREEN_SIZE) & input_buf & " o "
                buf = buf.Substring(buf.Length - SCREEN_SIZE, SCREEN_SIZE)
            Case DEC_MODE
                buf = Space(SCREEN_SIZE) & input_buf & " d "
                buf = buf.Substring(buf.Length - SCREEN_SIZE, SCREEN_SIZE)
            Case BIN_MODE
                buf = Space(SCREEN_SIZE) & input_buf & " b "
                buf = buf.Substring(buf.Length - SCREEN_SIZE, SCREEN_SIZE)
        End Select
        Calc_Form.Display.Text = buf
    End Sub

    ' Trim leading zeros from a string
    Private Function TrimZeros(ByVal buf As String) As String
        Dim i As Integer
        Dim b As String
        Dim is_neg As Boolean

        is_neg = False
        If buf.Substring(0, 1) = "-" Then
            i = 1
            is_neg = True
        End If

        Do
            b = buf.Substring(i, 1)
            i = i + 1
            If i >= buf.Length Then
                Exit Do
            End If
        Loop While b = "0"
        i = i - 1

        ' Leave at least the last zero
        If i = buf.Length Then
            i = i - 1
        End If

        If is_neg Then
            Return "-" & buf.Substring(i)
        Else
            Return buf.Substring(i)
        End If
    End Function

    Private Sub DoFlags(ByVal f As BigInt)
        If f.CarryBit() Then
            carry_bit = True
            DispFlags(DFLAG_CARRYBIT, True)
        Else
            carry_bit = False
            DispFlags(DFLAG_CARRYBIT, False)
        End If
        If f.Overflow() Then
            overflow = True
            DispFlags(DFLAG_OVERFLOW, True)
        Else
            overflow = False
            DispFlags(DFLAG_OVERFLOW, False)
        End If
    End Sub

    ' Display the status flags in the display
    Public Sub DispFlags(ByVal flag As Integer, ByVal state As Boolean)
        Select Case flag
            Case DFLAG_FKEY
                If state Then
                    Calc_Form.FLAG_F.Text = "f"
                Else
                    Calc_Form.FLAG_F.Text = ""
                End If
            Case DFLAG_GKEY
                If state Then
                    Calc_Form.FLAG_G.Text = "g"
                Else
                    Calc_Form.FLAG_G.Text = ""
                End If
            Case DFLAG_CARRYBIT
                If state Then
                    Calc_Form.FLAG_CARRYBIT.Text = "C"
                    Calc_Form.MenuO_Flag5.Checked = True
                Else
                    Calc_Form.FLAG_CARRYBIT.Text = ""
                    Calc_Form.MenuO_Flag4.Checked = False
                End If
            Case DFLAG_OVERFLOW
                If state Then
                    Calc_Form.FLAG_OVERFLOW.Text = "G"
                    Calc_Form.MenuO_Flag5.Checked = True
                Else
                    Calc_Form.FLAG_OVERFLOW.Text = ""
                    Calc_Form.MenuO_Flag5.Checked = False
                End If
        End Select
    End Sub

    ' *
    ' * Bunch of Utility Functions
    ' *

    ' Convert all Double variables to BigInt
    Private Sub AllTo64()
        Dim i As Integer

        For i = 0 To REG_SIZE - 1
            reg64(i) = BigInt.Assign(reg(i), word_size)
        Next
        For i = 0 To STACK_SIZE - 1
            stack64(i) = BigInt.Assign(stack(i), word_size)
        Next
        lastx64 = BigInt.Assign(lastx, word_size)
    End Sub

    ' Convert all BigInt variables to Doubles
    Private Sub AllFrom64()
        Dim i As Integer

        For i = 0 To REG_SIZE - 1
            reg(i) = reg64(i).ToDouble()
        Next
        For i = 0 To STACK_SIZE - 1
            stack(i) = stack64(i).ToDouble()
        Next
        lastx = lastx64.ToDouble()
    End Sub

    ' Resize all of the BigInt variables
    Public Sub ResizeAll(ByVal size As Integer)
        Dim i As Integer

        For i = 0 To REG_SIZE - 1
            reg64(i).SetSize(size)
        Next
        For i = 0 To STACK_SIZE - 1
            stack64(i).SetSize(size)
        Next
        lastx64.SetSize(size)

        sign_mask = BigInt.MakeSignMask(size, size)
    End Sub

End Module
